# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging

from gitlab.const import AccessLevel

from .base import BaseImporter
from .objects import GitLabObjectHandle

logger = logging.getLogger(__name__)


class LabelsImporter(BaseImporter):
    """Takes care of Project and Group labels."""

    def create_object_label(self, username, obj_handle, data):
        self.ensure_role(username, AccessLevel.REPORTER, obj_handle)
        api, extra = self.gitlab_user_api(username)
        return obj_handle.traverse(api).labels.create(
            self.normalize_payload(data),
            **extra
        )

    def create_project_label(self, username, project_id, data):
        return self.create_object_label(
            username,
            GitLabObjectHandle('project', project_id),
            data)

    def create_group_label(self, username, group_id, data):
        return self.create_object_label(username,
                                        GitLabObjectHandle('group', group_id),
                                        data)

    def ensure_object_labels(self, query_username, obj_handle, labels):
        """Create only missing project labels.

        :param query_username: the GitLab username of the user doing the
          query to list existing label.
        :param labels: a list of :class:`dict` suitable for
          :meth:`create_project_label` with an extra `username` value.
        """
        existing = set(self.gl_iterate_label_names(query_username, obj_handle))
        for label_data in labels:
            label_name = label_data['name']
            if label_name in existing:
                continue
            self.create_object_label(label_data.pop('username'),
                                     obj_handle,
                                     label_data)
            existing.add(label_name)

    def ensure_group_labels(self, query_username, group_id, labels):
        return self.ensure_object_labels(
            query_username, GitLabObjectHandle('group', group_id), labels)

    def ensure_project_labels(self, query_username, project_id, labels):
        return self.ensure_object_labels(
            query_username, GitLabObjectHandle('project', project_id), labels)

    def gl_iterate_label_names(self, username, obj_handle):
        api, extra = self.gitlab_user_api(username)
        # TODO in case of project, does this include group labels?
        manager = obj_handle.traverse(api).labels
        return (lab.name for lab in manager.list(iterator=True, **extra))



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from datetime import (  # noqa: F401 some are used in doctests only
    datetime,
    timezone,
    timedelta,
)


def dt_to_api(dt: datetime):
    """Convert an incoming datetime object to the string format of the API.

    The expected format is just iso-8601. This helper function makes sure
    that timezone information is present by resolving "naive" class:`datetime`
    objects as being UTC.

    As GitLab API often provides ISO-8601 with the `Z` token to mean UTC,
    we do the same.

    TODO consider as an alternative to raise ValueError from the onset.
    could be especially useful when we give importers a dry-run mode.

    >>> dt_to_api(datetime(2002, 4, 7, 11, 20, 0))
    '2002-04-07T11:20:00Z'
    >>> dt_to_api(datetime(2002, 4, 7, 11, 20, 0,
    ...     tzinfo=timezone(timedelta(hours=2))))
    '2002-04-07T11:20:00+02:00'
    """
    if dt.tzinfo is None:
        return dt.isoformat() + 'Z'
    return dt.isoformat()


def api_to_dt(iso: str):
    """Convert, understanding the 'Z' token of ISO 8601.

    >>> api_to_dt('2003-04-07T11:20:00Z')
    datetime.datetime(2003, 4, 7, 11, 20, tzinfo=datetime.timezone.utc)
    """
    if iso.endswith('Z'):
        iso = iso[:-1] + '+00:00'
    return datetime.fromisoformat(iso)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from attr import (
    define,
    Factory,
)
from datetime import datetime
import logging
from typing import (
    Any,
    Callable,
)

from gitlab.const import AccessLevel
from gitlab.exceptions import GitlabGetError

from .api import (
    GitLabMultipleAPI,
)
from .datetime import dt_to_api
from .objects import GitLabObjectHandle
logger = logging.getLogger(__name__)


ACCESS_LEVELS_BY_VALUE = {level.value: name
                          for name, level in AccessLevel.__members__.items()}


@define(slots=False)
class BaseImporter:
    """Abstract class taking care of common concerns

    It encapsulates the :class:`GitLabMultipleAPI` class to manage
    API connections for the relevant GitLab users.

    Concrete importers (issues, merge requests) are supposed to
    subclass this base class, but export/import applicative code is
    supposed to instantiate concrete importers rather than subclass again.

    All applicative notions (users, roles, objects) handled by this class
    and its subclasses supplied in `heptapod_api_import` are relevant to
    GitLab or Heptapod. They have no representation nor awaraness of the
    original system corresponding notions (notably users).
    """

    gitlab_multiple_api: GitLabMultipleAPI
    """Provides GitLab API access as the appropriate users."""

    roles_manager: Callable
    """A function providing the GitLab user for role management of an object.

    It must take a unique argument (a :class:`GitLabObjectHandler` and return
    a username.
    """

    roles_cache: dict = Factory(dict)
    """Used to remember access levels of users

    This spares lots of API calls.

    Keys are :class:`GitLabObjectHandle` instances, values are dicts
    whose keys are usernames and AccessLevel values.

    TODO If memory becomes an issue, we can replace this with an LRU cache.
    """

    comma_lists: Any = ('labels, ')
    """Control payload lists ro serialize as comma-separated strings.

    See :meth:`normalize_payload`.
    """

    @property
    def admin_api(self):
        return self.gitlab_multiple_api.gitlab.admin_api

    def gitlab_user_api(self, *args, **kwargs):
        """See :meth:`GitLabApiManager.user_api`."""
        return self.gitlab_multiple_api.user_api(*args, **kwargs)

    def get_role(self, username, obj: GitLabObjectHandle):
        """Get the role of username on the given Group or Project.

        Inheritance is taken into account.
        """
        cache = self.roles_cache.setdefault(obj, {})
        try:
            return cache[username]
        except KeyError:
            pass

        as_member = self.gl_user_as_member(username, obj)
        role = None if as_member is None else as_member.access_level
        cache[username] = role
        return role

    def gl_user_as_member(self, username, objh: GitLabObjectHandle):
        api, extra = self.gitlab_user_api(self.roles_manager(objh))
        user_id = self.gitlab_multiple_api.user_id(username)

        obj = objh.traverse(api, **extra)
        try:
            # Using `members_all` means that inherited members are included,
            # see, e.g., https://python-gitlab.readthedocs.io/en/stable/gl_objects/projects.html  # noqa: E501
            as_member = self.gl_members_get(obj.members_all, user_id, **extra)
            return as_member
        except GitlabGetError as exc:
            if exc.response_code != 404:
                raise
            return None

    def gl_members_get(self, obj_members, user_id, **extra):
        """Singled out for testing convenience."""
        return obj_members.get(user_id, **extra)

    def ensure_role(self, username, role, objh: GitLabObjectHandle):
        cache = self.roles_cache.setdefault(objh, {})
        existing_role = cache.get(username)
        if existing_role is not None and existing_role >= role:
            return

        as_member = self.gl_user_as_member(username, objh)
        if as_member is not None and as_member.access_level >= role:
            return

        api, extra = self.gitlab_user_api(self.roles_manager(objh))
        user_id = self.gitlab_multiple_api.user_id(username)
        obj = objh.traverse(api, **extra)
        if as_member is None:
            logger.info("Adding user %s to direct members of %s with "
                        "%s access level", username, objh, role.name)
            obj.members.create(dict(user_id=user_id, access_level=role),
                               **extra)
        else:
            logger.info("Bumping access level of user %r on %s "
                        "from %s to %s", username, objh,
                        ACCESS_LEVELS_BY_VALUE[as_member.access_level],
                        role.name)
            as_direct_member = obj.members.get(user_id, **extra)
            as_direct_member.access_level = role
            as_direct_member.save(**extra)

        cache[username] = role

    def normalize_payload(self, data):
        """Apply frequent conversions.

        - all datetime objects are converted to the expected ISO 8601 format
        - some lists (or tuples) are converted to comma-separated values,
          according to the :attr:`comma_lists` attribute.
        """
        normalized = {}
        for k, v in data.items():
            if isinstance(v, datetime):
                normalized[k] = dt_to_api(v)
            elif k in self.comma_lists:
                # in case of labels, that seem to be already handled by
                # python-gitlab
                normalized[k] = ','.join(v)
            else:
                normalized[k] = v
        return normalized



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from attr import define
from datetime import (
    datetime,
    time as dt_time,
    timedelta,
)
import logging
from typing import (
    Optional,
)

from gitlab import Gitlab as GitLabAPI

logger = logging.getLogger(__name__)

# Running hour of GitLab token expiry daily worker
TWO_AM_UTC = dt_time(hour=2)


class GitLabUserError(RuntimeError):
    """Base exception class for cases when an user cannot be leveraged.

    :attr:`args` is a singleton, made of the "username"
    """


class GitLabUserNotFound(GitLabUserError):
    """Raised when search is possible, but user was not found."""


class GitLabUserUnknown(GitLabUserError):
    """Raised when user is not known and search cannot be performed.

    Typically for cases where and admin token is not available.
    """


class GitLabUserNoToken(GitLabUserError):
    """Raised when no token is available for the user.

    Implicitely does not apply to implementations that are able to
    create tokens or to use the `sudo` feature of the GitLab API
    """


@define
class GitLab:
    url: str
    admin_api: Optional[GitLabAPI] = None
    admin_id: Optional[int] = None
    admin_username: Optional[str] = None

    def init_admin(self, token):
        api = self.admin_api = GitLabAPI(self.url, token)
        api.auth()
        self.admin_username = api.user.username
        self.admin_id = api.user.id
        logger.info("GitLab admin API connection checked (username=%r)",
                    self.admin_username)


class GitLabMultipleAPI:
    """Provide API access for GitLab users.

    Abstract base class to clarify the expected methods.

    Depending on implementation, the :attr:`users` attribute is used
    to authenticate or not (perhaps loaded from a file). In any case,
    it is used to carry interesting information about the user and can
    be used by downstream code to store their own interesting information.
    """

    def __init__(self, gitlab, users=None):
        self.gitlab = gitlab
        self.users = {} if users is None else users

    def user_api(self, username, utc_now=None):
        """Return an API connector authenticated as the given user.

        :param username: the "username" in the sens of GitLab, hence the
           login name / nickname, not the full name.
        :param utc_now: useful for tests of concrete classes if token
           expiration is to be taken into account.
        :return: a :class:`GitlabAPI` instance, and the extra parameters to
           pass when using it)
        :raises: GitLabUserError if impossible to act as the given user.
        """
        raise NotImplementedError  # pragma no cover

    def user_id(self, username):
        """Return the numeric id of the given user.

        We may in the future make it the handle used throughout by
        Heptapod API Import, in which case we'll probably have a ``username``
        method instead.
        """
        raise NotImplementedError  # pragma no cover


class GitLabPersonalTokensAPI(GitLabMultipleAPI):
    """Provide API access for GitLab users using Personal Tokens

    The personal tokens are expected to be found in the :attr:`users`
    attribute.

    Warnings can be issued in cases of close token expiration.
    """

    token_renewal_margin = timedelta(days=1)

    def user_api(self, username, utc_now=None):
        """See :meth:`GitlabMultipleAPI.user_api`
        """
        info = self.users.get(username)
        if info is None:
            raise GitLabUserUnknown(username)

        if utc_now is None:
            utc_now = datetime.utcnow()

        auth = info.get('auth')
        if auth is None:
            raise GitLabUserNoToken(username)

        exp = auth.get('token_expiration_date')
        # expiration can be unknown. This is the case of personal tokens.
        # TODO retrieve personal token information, but this can be
        # tricky (need the token name)
        if exp is not None and (exp - utc_now) < self.token_renewal_margin:
            logger.warning(
                "Access token for user %r will expire "
                "soon (in %d minutes).",
                username,
                (auth['token_expiration_date'] - utc_now
                 ).total_seconds() / 60,
            )

        api = info.get('api')
        if api is not None:
            return api, {}

        token = auth.get('token')
        if token is None:
            raise GitLabUserNoToken(username)

        auth['type'] = 'personal-token'
        api = info['api'] = GitLabAPI(self.gitlab.url, token)
        info['id'] = self.gl_current_user_id(api)
        return api, {}

    def user_id(self, username):
        self.user_api(username)
        return self.users[username]['id']

    def gl_current_user_id(self, api):
        """Wrapper method for easy monkey-patching in unit tests."""
        api.auth()
        return api.user.id


class GitLabSudoAPI(GitLabMultipleAPI):
    """Perform actions using the `sudo` feature of the GitLab API.

    The :attr:`gitlab` attribute is expected to be initialized with
    an admin API before any attempt at retrieving a regular user API.
    """

    @property
    def admin_api(self):
        return self.gitlab.admin_api

    def user_api(self, username):
        """Return API object and additional kwargs.

        The additional kwargs have to be used in API calls.
        """
        self.check_load_user(username)
        return self.gitlab.admin_api, dict(sudo=username)

    def user_id(self, username):
        return self.check_load_user(username)['id']

    def check_load_user(self, username):
        info = self.users.get(username)
        if info is not None:
            return info
        return self.load_user(username)

    def load_user(self, username):
        found = self.admin_api.users.list(username=username)
        if not found:
            raise GitLabUserNotFound(username)

        user = found[0]
        user_info = dict(id=user.id, auth=dict(type='sudo'))
        self.users[username] = user_info
        return user_info



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging

from gitlab.const import AccessLevel

from .base import BaseImporter
from .objects import GitLabObjectHandle

logger = logging.getLogger(__name__)


class IssuesImporter(BaseImporter):
    """Takes care of issues and associated objects.

    Applicative code should only need to use the :methd:`import_issues`
    method.
    """

    def __init__(self, *a, **kw):
        super(IssuesImporter, self).__init__(*a, **kw)
        self.saved_attrs = {}

    def start(self, project_id):
        """TODO make this a context manager?"""
        api, extra = self.gitlab_user_api('root')  # TODO object_owner
        project = api.projects.get(project_id, **extra)
        self.saved_attrs[GitLabObjectHandle('project', project_id)] = dict(
            emails_disabled=project.emails_disabled)
        project.heptapod_api_import = 'heptapod_api'
        project.emails_disabled = True
        project.save(**extra)

    def end(self, project_id):
        api, extra = self.gitlab_user_api('root')  # TODO object_owner
        project = api.projects.get(project_id, **extra)
        project.heptapod_api_import = 'heptapod_api_done'
        saved = self.saved_attrs[GitLabObjectHandle('project', project_id)]
        project.emails_disabled = saved['emails_disabled']
        project.save(**extra)

    def project_api(self, username, project_id):
        """Shortcut for readability."""
        api, extra = self.gitlab_user_api(username)
        return api.projects.get(project_id, lazy=True), extra

    def issues_api(self, username, project_id):
        """Shortcut for readability."""
        project, extra = self.project_api(username, project_id)
        return project.issues, extra

    def create_issue(self, data):
        """Create just the issue itself."""
        project_id = data.pop('project_id')
        username = data.pop('username')
        data = self.normalize_payload(data)

        title = data['title']
        if len(title) > 255:
            data['title'] = title[:200] + '...'
            descr = data.get('description', '').strip()
            if descr:
                data['description'] = title + '\n\n' + descr
            else:
                data['description'] = title

        if not data.get('description'):
            # GitLab does not accept empty issue descriptions
            data['description'] = '(empty description)'

        return self.gl_create_issue(project_id, username, data)

    def gl_create_issue(self, project_id, username, data):

        issues, extra = self.issues_api(username, project_id)
        # TODO need the admin_issue permission for some operations
        # and the `set_issue_metadata` for labels, milestones etc.
        # cf app/policies/issue_policy.rb
        return issues.create(data, **extra)

    def ensure_user_can_file(self, username, project_id):
        """Give the minimal permissions so that user can file issue, comment.
        """
        # TODO unless project is public
        self.ensure_role(username, AccessLevel.REPORTER,
                         GitLabObjectHandle('project', project_id)
                         )

    def create_note(self, issue, data):
        username = data.pop('username')
        self.ensure_user_can_file(username, issue.project_id)
        # TODO unless project is public
        data = self.normalize_payload(data)
        if not data.get('body'):
            # GitLab does not accept empty body:
            #     400 Bad request - Note {:note=>["can't be blank"]}
            data['body'] = '(empty message)'
        return self.gl_create_note(username, issue, data)

    def gl_create_note(self, username, issue, data):
        issues, extra = self.issues_api(username, issue.project_id)
        return issues.get(issue.iid).notes.create(data, **extra)

    def usernames_to_ids(self, data, key, target_key):
        """Update data[key] in place to put ids instead of usernames.
        """
        value = data.pop(key, None)
        if value is None:
            return

        data[target_key] = tuple(self.gitlab_multiple_api.user_id(v)
                                 for v in value)

    def change_labels(self, issue, data):
        username = data.pop('username')
        issues, extra = self.issues_api(username, issue.project_id)
        as_user = issues.get(issue.iid, **extra)
        data = self.normalize_payload(data)
        as_user.add_labels = data['add_labels']
        as_user.remove_labels = data['remove_labels']
        dt = data.get('datetime')
        state_change = data.get('state_change')
        if state_change is not None:
            as_user.state_event = state_change
        if dt is not None:
            as_user.updated_at = dt
        assignee_ids = data.get('assignee_ids')
        if assignee_ids is not None:
            as_user.assignee_ids = assignee_ids

        as_user.save(**extra)

    def create_project_file(self, project_id, data):
        username = data.pop('creator')
        project, extra = self.project_api(username, project_id)
        return project.upload(data['file_name'], filepath=data['path'],
                              **extra)

    def import_issue(self, data, attachments_header='\n\n'):
        """Import an issue and its timeline of events.

        The events can be a new comment, changing labels, closing, reopening
        etc.

        Data dictionaries are as in the GitLab API PUT or POST body,
        with the following exceptions:

        - ``events`` (issue itself only): list of dict-like objects with:
          + ``type``: one of the implemented event types (see below)
          + ``data``: actual data associated to the event. They are
           subject to the rules explained here.
        - ``project_id`` (issue itself only): numeric ID (:class:`int`)
           of the project. Not part of the API body, because it is passed
          in the URL.
        - ``username`` (issue itself and all its events): used by the
          importer to act as the proper GitLab user. Not part of API body
          (derived from token headers).
        - all date/time values: have to be given in UTC as a
          :class:`datetime` instance without time zone.
        - ``labels`` (issue itself and all its events): iterable of label
          names
        - ``attachments``: list of dict-like objects with:
          + ``creator``: username of the attachment creator
          + ``path``: path on the local filesystem to data (TODO support also
            opended file-like object)
          + ``file_name``: name of the attached file
          + ``download_renderer``: template to use to render the attachment
            in issue description, will be given the projects file return
            dict, with ``url`` and ``markdown`` keys, see
              https://python-gitlab.readthedocs.io/en/stable/gl_objects/projects.html#file-uploads
            Together with ``attachments_header``, can be used, e.g., to
            build up a table.

        Events and their specificites:

        - ``comment``: an Issue Note is added to the issue
        - ``change_labels``: an update of the issue is made to add labels.
          This event has specific data fields: `add_labels`, `remove_labels`
          (both iterables of label names) and `datetime`.

        Note: the "event" terminology is specific to Heptapod API Importer.

        GitLab API References:

        - Issues: https://docs.gitlab.com/ce/api/issues.html
        - Notes (comments): https://docs.gitlab.com/ce/api/notes.html
        """
        # TODO decide whether we want to grant or use the owner API
        # to add the labels
        project_id = data['project_id']
        self.ensure_user_can_file(data['username'], project_id)
        events = data.pop('events')
        attachments = data.pop('attachments', ())
        if attachments:
            data['description'] += attachments_header
        for att in attachments:
            self.ensure_user_can_file(att['creator'], project_id)
            render_args = self.create_project_file(data['project_id'], att)
            data['description'] += att['download_renderer'].format(
                **render_args)
        self.usernames_to_ids(data, 'assignees', 'assignee_ids')

        issue = self.create_issue(data)

        for event in events:
            event_data = event['data']
            self.usernames_to_ids(event_data, 'assignees', 'assignee_ids')
            if event['type'] == 'comment':
                self.create_note(issue, event_data)
            elif event['type'] == 'change_labels':
                self.change_labels(issue, event_data)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import attr


@attr.define(frozen=True)
class GitLabObjectHandle:
    """A lightweight object specifier.

    It is much smaller than a full python-gitlab object and is purely
    descriptive.
    """
    type: str
    """The object type.

    In general, this is the singular form, in lower case.
    Current supported types are ``project`` and ``group``.
    """
    id: int
    """The object numeric id."""

    def __str__(self):
        """Nice representation for logging etc.

        The adopted format is similar to the `repr()` of python-gitlab
        objects::

          >>> str(GitLabObjectHandle(type='project', id=87))
          '[Project id:87]'

        Square brackets instead of angled to make the difference visible.
        """
        return f'[{self.type.capitalize()} id:{self.id}]'

    def retrieve(self, api, **extra):
        """Retrieve actual python-gitlab object using the given API.
        """
        return getattr(api, self.type + 's').get(self.id, **extra)

    def traverse(self, api, **extra):
        """Perform the "lazy get of python-gitlab using the given API."""
        return self.retrieve(api, lazy=True, **extra)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import attr
import logging
import re

from typing import (
    Optional,
    Callable,
)
from gitlab.const import SearchScope
from gitlab.exceptions import (
    GitlabCreateError,
)
from .api import GitLabAPI
from .datetime import dt_to_api

logger = logging.getLogger(__name__)


class IdentityMapping:

    def get(self, k):
        return k


GITLAB_USERNAME_BASE_RX = re.compile(r'^\w+$')
GITLAB_USERNAME_FORBIDDEN_SUFFIXES = ('.', '.git', '.atom')
# hyphen in Python regexps character sets need no escaping if first or last:
GITLAB_USERNAME_FORBIDDEN_CHARS_RX = re.compile(r'[^a-zA-Z0-9_.-]')
EMAIL_AT_AND_FOLLOWING_RX = re.compile(r'@.*$')
LEADING_HYPHENS_RX = re.compile(r'^-+')
DECIMAL_NUMBER_RX = re.compile(r'^\d+')


def is_valid_username(username):
    """Tell wheteher the username respects GitLab standards.

    # From warnings emitted by GitLab API:
    # 'username': ["can contain only letters, digits, '_', '-' and '.'.
    # Cannot start with '-' or end in '.', '.git' or '.atom'."]}

    >>> is_valid_username('someone@something')
    False
    >>> is_valid_username('-foo')
    False
    >>> is_valid_username('some-foo_bar')
    True
    >>> is_valid_username('gitlab.git')
    False
    """
    return (GITLAB_USERNAME_FORBIDDEN_CHARS_RX.search(username) is None
            and not username.startswith('-')
            and not any(username.endswith(suffix)
                        for suffix in GITLAB_USERNAME_FORBIDDEN_SUFFIXES
                        )
            )


def sanitize_username(name):
    r"""Return a valid GitLab username derived from the provided one.

    If the provided one is valid, it is of course unchanged. Otherwise,
    the rules that GitLab applies in SSO signup are applied.

    From ``app/models/namespace.rb`` in GitLab 14.10.1, called from
    ``Gitlab::Auth::OAuth::User#user_attributes``::

      def clean_path(path)
        path = path.dup
        # Get the email username by removing everything after an `@` sign.
        path.gsub!(/@.*\z/,                "")
        # Remove everything that's not in the list of allowed characters.
        path.gsub!(/[^a-zA-Z0-9_\-\.]/,    "")
        # Remove trailing violations ('.atom', '.git', or '.')
        loop do
          orig = path
          PATH_TRAILING_VIOLATIONS.each { |ext| path = path.chomp(ext) }
          break if orig == path
        end

        # Remove leading violations ('-')
        path.gsub!(/\A\-+/, "")

        # Users with the great usernames of "." or ".." would end up with
        # a blank username.
        # Work around that by setting their username to "blank", followed by
        # a counter.
        path = "blank" if path.blank?

        uniquify = Uniquify.new
        uniquify.string(path) { |s| Namespace.find_by_path_or_name(s) }
      end

    This function does the same, expcept for the unicity counter at the end,
    which will require round-trips to the GitLab instance.


    Doctests inspired by `spec/models/namespace_spec.rb` at GitLab 14.10.1::

      >>> sanitize_username("-john+gitlab-ETC%.git@gmail.com")
      'johngitlab-ETC'
      >>> sanitize_username("--%+--valid_*&%name=.git.%.atom.atom.@email.com")
      'valid_name'

    More tests::

      >>> sanitize_username('.git')
      'blank'

    Numeric usernames are problematic with the APU, especially when using
    the ``sudo`` feature, because they are interpreted as user ids. Hence we
    change them:

       >>> sanitize_username('1234567')
       '1234567_'

    """
    # Handling multiple lines: it is very unlikely that the system we're
    # exporting users from would accept line breaks in nicknames.
    # Still, let's guard against it.
    name = name.splitlines()[0]
    # (If supplied username looks like an email address),
    # get the email username by removing everything after an `@` sign.
    # (including the `@`)
    name = EMAIL_AT_AND_FOLLOWING_RX.sub('', name)

    # Remove everything that's not in the list of allowed characters.
    name = GITLAB_USERNAME_FORBIDDEN_CHARS_RX.sub('', name)
    while True:
        orig = name
        for suffix in GITLAB_USERNAME_FORBIDDEN_SUFFIXES:
            if name.endswith(suffix):
                name = orig[:-len(suffix)]
        if name == orig:
            break

    # Remove leading violations ('-')
    name = LEADING_HYPHENS_RX.sub('', name)

    # Users with the great usernames of "." or ".." would end up with a
    # blank username.
    # Work around that by setting their username to "blank", followed by a
    # counter.
    # (the counter part is left to the caller)
    if not name:
        name = "blank"

    # Pure numeric usernames are problematic with the API, because
    # the `sudo` feature interprets them as ids, not usernames
    if DECIMAL_NUMBER_RX.match(name):
        name += '_'

    return name


@attr.define
class EmailUserMapping:
    """A user mapping whose keys are email addresses, based on GitLab lookups.

    Relies on the admin connection provided by :attr:`admin_api`.
    """
    admin_api: GitLabAPI
    mapping: dict = attr.field(factory=dict)

    create_missing: bool = False
    """If ``True`` missing users will be automatically created.

    For the creation to work, the  :attr:`src_user_exporter` attribute
    must also be set.
    """

    src_user_exporter: Optional[Callable] = None
    """Callback to extract the necessary data to create missing GitLab users.

    The callback takes a unique :class:`str` argument (the user email)

    Callback return
    ~~~~~~~~~~~~~~~

    The returned value must be a :class:`dict`.

    Required values:

    - `nickname`: the user short handle, usable in login or in `@` mentions.
       Example: 'gracinet'
    - `fullname`: the complete name, e.g., 'Georges Racinet'

    Default values

    - `force_random_password`: set to `True` if `password` and `reset_password`
      are missing. The create user will have to use the reset facility at
      first login.
    - `confirmed_at`: if not set, the current time is automatically used.

    Ignored values:

    - `email`: the input email will be used as primary email. This can change
      in the future if there are use cases for secondary emails.

    Unsettable values:

    - `confirmed_at`, `created_at`: ignored by GitLab itself. Applications
      are advised to use the ``note`` or ``bio`` text attributes to record
      the original date of creation for lack of better solution.

    Other key/value pairs will be passed as-is to the
    `GitLab API <https://docs.gitlab.com/ce/api/users.html#user-creation>`_

    Note: the returned :class:`dict` will be converted in-place, hence
    after-the-fact inspection is possible.
    """

    testing_mode: bool = False
    """Put created addresses under the :attr:`testing_address_domain` domain.

    If ``True``, automatic user creation will be made by postfixing the
    address with the value of :addr:`testing_address_domain`, so that, e.g,
    ``foo@example.com`` becomes ``foo@example.com.heptapod.test``.

    This is meant for testing instances, to be 100% sure that a test run
    wont end up sending confusing emails to the actual users.

    Addresses that are found in the database are unaffected: if a user
    has the requested address, it will be found. Similarly, if
    ``foo@example.com.heptapod.test`` is present in the database, its user
    will match ``foo@example.com``, so that resuming an import that has
    performed some automatic creations works straight.
    """

    testing_address_domain: str = 'heptapod.test'
    """See :attr:`testing_mode."""

    @property
    def testing_address_suffix(self):
        return '.' + self.testing_address_domain

    def normalize_email(self, email):
        """Normalization as GitLab does.

        Excerpts from Rails app (GitLab 14.10.1) lib/api/user.rb::

            conflict!('Email has already been taken') if User
              .by_any_email(user.email.downcase)
              .any?

        and also (TODO not sure where exactly it matters, but we may need
        to implement something alike)
        app/services/users/update_canonical_email_service.rb::

            def canonicalize_email
              email = user.email

              portions = email.split('@')
              username = portions.shift
              rest = portions.join

              regex = Regexp.union(INCLUDED_DOMAINS_PATTERN)
              return unless regex.match?(rest)

              no_dots = username.tr('.', '')
              before_plus = no_dots.split('+')[0]
              "#{before_plus}@#{rest}"
            end
        """
        return email.lower()

    def get(self, email):
        gl_email = self.normalize_email(email)

        # even in testing mode, it could be that the user exists with
        # the given address (preloaded testing instance)
        username = self.simple_fetch(gl_email)
        if username is not None:
            return username

        if self.testing_mode:
            gl_email += self.testing_address_suffix
            username = self.simple_fetch(gl_email)

        if username is None and self.create_missing:
            username = self.create_user(gl_email, input_email=email)

        return username

    def simple_fetch(self, email):
        gl_username = self.mapping.get(email)
        if gl_username is not None:
            return gl_username

        gl_username = self.gl_exact_search(email)
        if gl_username is not None:
            self.mapping[email] = gl_username
        return gl_username

    def gl_exact_search(self, email):
        """Use the fuzzy GitLab search API and postfilter for exact matches.

        :returns: a :class:`User` instance or `None` if no match.

        Only *confirmed* email addresses are taken into account.

        GitLab API allows to search users by email only with the general
        search (Intended for fulltext / UI convenience) that does search on
        primary and secondary emails, but also takes unrelated
        fields into account (e.g., ``username``, or whatever will feel useful
        for humans in the future).

        This method adds a layer of check to eliminate wrong (or unconfirmed
        results).

        Whatever hops we have to take to perform this search, the results
        eventually come from the `emails` table, which map email adresses
        to users without ambiguity because:

        - it has a unicity constraint on the email address itself
        - it has a foreign key constraint towards the users table
        """
        api = self.admin_api
        fuzzy = api.search(SearchScope.USERS, email)
        for info in fuzzy:
            user = api.users.get(info['id'])
            for user_email in user.emails.list():
                if user_email.email == email:
                    # depending on GitLab version, unconfirmed emails can
                    # be returned or not, whence the pragma.
                    if user_email.confirmed_at is None:  # pragma no cover
                        return None
                    return user.username

    def create_user(self, gl_email, input_email):
        data = self.src_user_exporter(input_email)
        self.convert_user_data(gl_email, data)
        input_username = data.pop('input_username')
        attempt_username = data['username']
        count = 1
        while count < 10:
            try:
                self.gl_create_user(data)
                break
            except GitlabCreateError as exc:
                err_msg = exc.error_message.lower().strip()
                if (exc.response_code != 409
                        or err_msg != 'username has already been taken'):
                    raise
                count += 1
                data['username'] = f'{attempt_username}_{count}'

        final_username = data['username']
        self.mapping[gl_email] = final_username
        if final_username != input_username:
            logger.warning("Final username associated with email address %r "
                           "is %r instead of the exported %r, to respect "
                           "GitLab validity rules. The user can change it "
                           "afterwards if the GitLab instance configuration "
                           "allows it.",
                           input_email, final_username, input_username)
        return final_username

    def convert_user_data(self, email, data):
        """Convert user data to match GitLab schema."""
        input_username = data.pop('nickname')
        data['input_username'] = input_username
        data['username'] = sanitize_username(input_username)
        data['name'] = data.pop('fullname')
        data['email'] = email
        confirmed_at = data.pop('confirmed_at', None)
        if confirmed_at is not None:
            data['confirmed_at'] = dt_to_api(confirmed_at)
        if not data.get('password') and not data.get('reset_password'):
            data['force_random_password'] = True

        # If skip_confirmation value below changes, then the logic to match
        # users by confirmed email addressses only (necessary for security)
        # also needs to be adapted.
        data['skip_confirmation'] = True
        return data

    def gl_create_user(self, data):
        logger.info("Creating user %r", data)
        return self.admin_api.users.create(data)



from pathlib import Path
from setuptools import setup

MAIN_PACKAGE = 'heptapod_api_import'
VERSION_FILE = 'VERSION'
REQUIREMENTS_FILE = 'install-requirements.txt'


scripts = {  # name to (module in main package, function)
}


setup(
    name='heptapod-api-import',
    version=Path(MAIN_PACKAGE, VERSION_FILE).read_text().strip(),
    author='Georges Racinet',
    author_email='georges.racinet@octobus.net',
    url='https://foss.heptapod.net/heptapod/heptapod-api-import',
    description="Toolbox library for using the GitLab/Heptapod API for "
    "import use cases",
    long_description=Path('README.md').read_text(),
    long_description_content_type="text/markdown",
    keywords='hg mercurial git heptapod gitlab',
    license='GPL3+',
    # do not use find_packages, as it could recurse into the Git and
    # Mercurial repositories
    packages=[MAIN_PACKAGE],
    package_data={MAIN_PACKAGE: [VERSION_FILE]},
    entry_points=dict(
        console_scripts=[
            '{name}={pkg}.{mod}:{fun}'.format(
                pkg=MAIN_PACKAGE, name=name, mod=mod, fun=fun)
            for name, (mod, fun) in scripts.items()],
    ),
    install_requires=Path(REQUIREMENTS_FILE).read_text().splitlines(),
)
